/* ###################################################################
**     Filename    : mqx_tasks.c
**     Project     : RemotePowerControl
**     Processor   : MK64FN1M0VLQ12
**     Component   : Events
**     Version     : Driver 01.00
**     Compiler    : GNU C Compiler
**     Date/Time   : 2014-12-05, 22:39, # CodeGen: 2
**     Abstract    :
**         This is user's event module.
**         Put your event handler code here.
**     Settings    :
**     Contents    :
**         Task1_task - void Task1_task(void* task_init_data);
**
** ###################################################################*/
/*!
** @file mqx_tasks.c
** @version 01.00
** @brief
**         This is user's event module.
**         Put your event handler code here.
*/         
/*!
**  @addtogroup mqx_tasks_module mqx_tasks module documentation
**  @{
*/         
/* MODULE mqx_tasks */

#include "Cpu.h"
#include "Events.h"
#include "mqx_tasks.h"
#include "config.h"
#include <nio_tfs.h>
#include <shell.h>

#ifdef __cplusplus
extern "C" {
#endif 


/* User includes (#include below this line is not maintained by Processor Expert) */
#ifndef BSP_ENET_DEVICE_COUNT
#error Ethernet devices not defined!
#endif

#if BSP_ENET_DEVICE_COUNT == 0
#error No ethernet devices defined!
#endif

const SHELL_COMMAND_STRUCT Shell_commands[] = {
   /* RTCS commands */
   { "help",      Shell_help },
   { "ipconfig",  Shell_ipconfig },
   { "netstat",   Shell_netstat },
#if RTCSCFG_ENABLE_ICMP
   { "ping",      Shell_ping },
#endif
   { "?",         Shell_command_list },
   { NULL,        NULL }
};

/*
** ===================================================================
**     Event       :  Task1_task (module mqx_tasks)
**
**     Component   :  Task1 [MQX_KSDK_Task]
*/
/*!
**     @brief
**         MQX task routine. The routine is generated into mqx_tasks.c
**         file.
**     @param
**         task_init_data  - 
*/
/* ===================================================================*/
void Task1_task(uint32_t task_init_data)
{
    int32_t                                 error;
    uint32_t                                server[BSP_ENET_DEVICE_COUNT];
    extern const HTTPSRV_CGI_LINK_STRUCT    cgi_lnk_tbl[];
    extern const HTTPSRV_SSI_LINK_STRUCT ssi_lnk_tbl[];
    extern const NIO_TFS_DIR_ENTRY          nio_tfs_data[];
    HTTPSRV_PARAM_STRUCT                    params[BSP_ENET_DEVICE_COUNT];
    uint32_t                                ip_addr[BSP_ENET_DEVICE_COUNT];
    uint32_t                                i = 0;
    char*                                   indexes[BSP_ENET_DEVICE_COUNT];
    uint8_t                                 n_devices = BSP_ENET_DEVICE_COUNT;
    #if !HTTP_USE_ONE_SERVER
    uint8_t n_servers = BSP_ENET_DEVICE_COUNT;
    #else
    uint8_t n_servers = 1;
    #endif
    void                                 *ssl_context;

    /* Init RTCS */
    _RTCSPCB_init = 4;
    _RTCSPCB_grow = 2;
    _RTCSPCB_max = 8;

    _RTCSTASK_stacksize = 4500;

    error = RTCS_create();
    if (error != RTCS_OK)
    {
        printf("RTCS failed to initialize, error = 0x%X\n", error);
        _task_block();
    }

    _IP_forward = TRUE;


    /* Bind IP address for each interface */
    for (i = 0; (i < n_devices) && (n_devices != 0); i++)
    {
        _enet_address           address;
        uint32_t                 phy_addr;
        #if RTCSCFG_ENABLE_IP4
        IPCFG_IP_ADDRESS_DATA   ip_data;
        #endif
        phy_addr = i;
        ip_addr[i] = IPADDR(ENET_IPADDR_A,ENET_IPADDR_B,ENET_IPADDR_C+i,ENET_IPADDR_D);

        ENET_get_mac_address(phy_addr, ip_addr[i], address);

        /* Init ENET device */
        error = ipcfg_init_device (phy_addr, address);
        if (error != RTCS_OK)
        {
            printf("IPCFG: Device n.%d init failed. Error = 0x%X\n", i, error);
            _task_set_error(MQX_OK);
            n_devices--;
            i--;
            continue;
        }

        #if RTCSCFG_ENABLE_IP4
        ip_data.ip      = ip_addr[i];
        ip_data.mask    = ENET_IPMASK;
        ip_data.gateway = 0;
        /* Bind IPv4 address */
        error = ipcfg_bind_staticip (phy_addr, &ip_data);
        if (error != RTCS_OK)
        {
            printf("\nIPCFG: Failed to bind IP address. Error = 0x%X", error);
            _task_block();
        }
        #endif

        indexes[i] = (char*) _mem_alloc_zero(sizeof("\\index_x.html"));
        if (indexes[i] == NULL)
        {
            printf("\n Failed to allocate memory.");
            _task_block();
        }
    }

    NIO_TFS_INIT_DATA_STRUCT tfs_init = { nio_tfs_data };
    if (_nio_dev_install("tfs:", &nio_tfs_dev_fn, &tfs_init) == NULL)
    {
        printf("\n Failed to initialize Trivial File System");
        _task_block();
    }

    ssl_context = RTCS_ssl_init((RTCS_SSL_PARAMS_STRUCT *) &ssl_params);
    /* Start HTTP server on each interface */
    for (i = 0; i < n_servers; i++)
    {
        char *secure_string;
        _mem_zero(&params[i], sizeof(HTTPSRV_PARAM_STRUCT));
        params[i].af = HTTP_INET_AF;  //IPv4, IPv6 or from config.h

        #if RTCSCFG_ENABLE_IP6
        /* Set interface number here. Zero is any. */
        params[i].ipv6_scope_id = HTTP_SCOPE_ID;
        #endif

        sprintf(indexes[i], "\\index.html");
        params[i].root_dir = "tfs:\\web";
        params[i].index_page = indexes[i];
        params[i].auth_table = (HTTPSRV_AUTH_REALM_STRUCT*) auth_realms;
        secure_string = (ssl_context == NULL) ? " " : "S ";
        printf("Starting HTTP%sserver No.%d on IP", secure_string, i);
        /*
        ** If there is only one server listen on any IP address
        ** so address can change in runtime (DHCP etc.).
        ** Otherwise we will use static IP for server.
        */
        if (HTTP_INET_AF & AF_INET)
        {
            #if RTCSCFG_ENABLE_IP4
            if ((i == 0) && (n_servers == 1))
            {
                params[i].ipv4_address.s_addr = INADDR_ANY;
            }
            else
            {
                params[i].ipv4_address.s_addr = ip_addr[i];
            }
            /* Print active IPv4 address */
            printf(" %d.%d.%d.%d", IPBYTES(ip_addr[i]));
            #endif
        }
        if (HTTP_INET_AF & AF_INET6)
        {
            #if RTCSCFG_ENABLE_IP6
            IPCFG6_GET_ADDR_DATA data;
            char prn_addr6[RTCS_IP_ADDR_STR_SIZE];

            ipcfg6_get_addr(0, 0, &data);

            if(inet_ntop(AF_INET6, &(data.ip_addr), prn_addr6, sizeof(prn_addr6)))
            {
                /* Print active IPv6 address */
                printf("%s %s", (HTTP_INET_AF & AF_INET) ? " and" : "", prn_addr6);
            }
            #endif
        }
        printf(", port %d", HTTPSRV_PORT);
        /*
        ** Maximum default number of available sockets in RTCS is six.
        ** Maximum of two are used for HTTP server as listening sockets (one for IPv4 and one for IPv6).
        ** So we take two of sockets and use them for sessions.
        */
        params[i].max_ses = HTTPSRV_EXAMPLE_SESSIONS;
        params[i].cgi_lnk_tbl = (HTTPSRV_CGI_LINK_STRUCT *) cgi_lnk_tbl;
        params[i].ssi_lnk_tbl = (HTTPSRV_SSI_LINK_STRUCT *) ssi_lnk_tbl;
        params[i].plugins = &plugins;
        params[i].port = HTTPSRV_PORT;
        params[i].ssl_ctx = ssl_context;
        init_plugins();
        /* There are a lot of static data in CGIs so we have to set large enough stack for script handler */
        params[i].script_stack = HTTPSRV_EXAMPLE_CGI_STACK;
        server[i] = HTTPSRV_init(&params[i]);

        printf("...");
        if (server[i] == 0)
        {
            printf("[FAIL]\n");
        }
        else
        {
            printf("[OK]\n");
        }
    }

    /* Start Shell.*/
    for (;;)
    {
        /* Run the shell */
        Shell(Shell_commands, NULL);
        printf("Shell quit, restarting...\n");
    }
}

/* END mqx_tasks */

#ifdef __cplusplus
}  /* extern "C" */
#endif 

/*!
** @}
*/
/*
** ###################################################################
**
**     This file was created by Processor Expert 10.4 [05.10]
**     for the Freescale Kinetis series of microcontrollers.
**
** ###################################################################
*/
